@@ -1,7 +1,8 @@
 (******************************************************************************)
 (* OASIS: architecture for building OCaml libraries and applications          *)
 (*                                                                            *)
-(* Copyright (C) 2008-2010, OCamlCore SARL                                    *)
+(* Copyright (C) 2011-2013, Sylvain Le Gall                                   *)
+(* Copyright (C) 2008-2011, OCamlCore SARL                                    *)
 (*                                                                            *)
 (* This library is free software; you can redistribute it and/or modify it    *)
 (* under the terms of the GNU Lesser General Public License as published by   *)
@@ -26,6 +27,13 @@ open OASISUtils
 open OASISTypes
 open OASISString
 
+
+(* TODO: logic of this file is brittle, we update digest frequently and not
+ * always with the same content.
+ * Proposal: create functions add_string [`Header|`Body|`Footer] t str (side
+ * effect) and update digest into a single place.
+ *)
+
 type comment =
     {
       (** Return the string as a comment. *)
@@ -36,26 +44,33 @@ type comment =
       stop:      string;
     }
 
+
 type line = string
 
+
 type body =
   | NoBody
   | Body of line list
   | BodyWithDigest of Digest.t * line list
 
+
 type template =
     {
-      fn:      host_filename;
-      comment: comment;
-      header:  line list;
-      body:    body;
-      footer:  line list;
-      perm:    int;
+      fn:                    host_filename;
+      comment:               comment;
+      header:                line list;
+      body:                  body;
+      footer:                line list;
+      perm:                  int;
+      important:             bool;
+      disable_oasis_section: bool;
     }
 
+
 let start_msg = "OASIS_START"
 let stop_msg  = "OASIS_STOP"
 
+
 let comment cmt_beg cmt_end =
   let of_string =
       match cmt_end with
@@ -87,37 +102,29 @@ let comment cmt_beg cmt_end =
       stop      = of_string stop_msg;
     }
 
-let comment_ml =
-  comment "(*" (Some "*)")
-
-let comment_sh =
-  comment "#" None
-
-let comment_makefile =
-  comment_sh
-
-let comment_ocamlbuild =
-  comment_sh
-
-let comment_bat =
-  comment "rem" None
-
-let comment_meta =
-  comment_sh
 
+let comment_ml = comment "(*" (Some "*)")
+let comment_sh = comment "#" None
+let comment_makefile = comment_sh
+let comment_ocamlbuild = comment_sh
+let comment_bat = comment "rem" None
+let comment_meta = comment_sh
+let comment_markdown = comment "<!---" (Some "--->")
 
 let template_make fn comment header body footer =
   {
-    fn      = fn;
-    comment = comment;
-    header  = header;
-    body    = Body body;
-    footer  = footer;
-    perm    = 0o644;
+    fn                    = fn;
+    comment               = comment;
+    header                = header;
+    body                  = Body body;
+    footer                = footer;
+    perm                  = 0o644;
+    important             = false;
+    disable_oasis_section = false;
   }
 
 
-let template_of_string_list ~ctxt ~template fn comment lst =
+let template_of_string_list ~ctxt ~template ?(disable_oasis_section=false) fn comment lst =
 
   (* Convert a Digest.to_hex string back into Digest.t *)
   let digest_of_hex s =
@@ -147,7 +154,7 @@ let template_of_string_list ~ctxt ~template fn comment lst =
     match comment.to_string str with
       | Some str' ->
           begin
-            match tokenize ~tokens:["(";")";":"] str' with
+            match tokenize ~tokens:["("; ")"; ":"] str' with
               | ["DO"; "NOT"; "EDIT"; "("; "digest"; ":"; digest; ")"] ->
                   digest
               | _ ->
@@ -180,14 +187,14 @@ let template_of_string_list ~ctxt ~template fn comment lst =
         try
           let lst_header, tl =
             split_cond
-              (fun str ->
-                if not (is_start str) then
-                  begin
+              (if disable_oasis_section then
+                fun _ -> false
+              else
+                fun str ->
+                  if not (is_start str) then begin
                     debug ~ctxt "Not start: %s" str;
                     true
-                  end
-                else
-                  begin
+                  end else begin
                     debug ~ctxt "Start: %s" str;
                     false
                   end)
@@ -222,7 +229,10 @@ let template_of_string_list ~ctxt ~template fn comment lst =
                   lst_header, Body lst_body, lst_footer
 
         with Not_found ->
-          lst, NoBody, []
+          if disable_oasis_section then
+            [], Body [], []
+          else
+            lst, NoBody, []
 
   in
 
@@ -245,7 +255,7 @@ let template_of_string_list ~ctxt ~template fn comment lst =
     {res with body = body}
 
 
-let template_of_file ~template fn comment =
+let template_of_file ~template disable_oasis_section fn comment =
  let lst =
    let chn_in =
      open_in_bin fn
@@ -264,7 +274,7 @@ let template_of_file ~template fn comment =
      close_in chn_in;
      List.rev !lst
  in
-   template_of_string_list ~template fn comment lst
+   template_of_string_list ~template ~disable_oasis_section fn comment lst
 
 
 let template_of_mlfile fn header body footer  =
@@ -293,7 +303,7 @@ let template_of_mlfile fn header body footer  =
     let found = ref false in
     let extract_line_modifier str =
       if starts_with "#" (trim str) then
-        try 
+        try
           match tokenize_genlex str with
             | [Genlex.Ident "#"; Genlex.Int _; Genlex.String fn] ->
                 fn
@@ -324,7 +334,7 @@ let template_of_mlfile fn header body footer  =
 
            with Not_found ->
              line)
-        (split_newline ~trim:false str)
+        (split_newline ~do_trim:false str)
     in
       !found, (String.concat "\n" lst)
   in
@@ -383,6 +393,7 @@ let template_of_mlfile fn header body footer  =
       body
       footer
 
+
 let digest_update t =
   {t with
        body =
@@ -390,9 +401,9 @@ let digest_update t =
            | NoBody -> NoBody
            | BodyWithDigest (_, lst)
            | Body lst ->
-               BodyWithDigest
-                 (Digest.string (String.concat "\n" lst),
-                  lst)}
+               let d = Digest.string (String.concat "\n" lst) in
+                 BodyWithDigest(d, lst)}
+
 
 let digest_check t =
   let t' = digest_update t in
@@ -402,69 +413,75 @@ let digest_check t =
       | _, _ ->
           true
 
+
 let merge t_org t_new =
-  {t_new with
-       header = t_org.header;
-       body   =
-         (if t_org.body = NoBody then
-            t_org.body
-          else
-            t_new.body);
-       footer = t_org.footer}
+  digest_update
+    {t_new with
+         header = t_org.header;
+         body   =
+           (if t_org.body = NoBody then
+              t_org.body
+            else
+              t_new.body);
+         footer = t_org.footer}
 
-let to_file t =
+
+let to_string_list t =
   (* Be sure that digest match body content *)
-  let t =
-    digest_update t
-  in
-
-  (* Write body, header and footer to output file. Separate
+  let t = digest_update t in
+  (* Create header, body and footer in a list. Separate
    * each part by appropriate comment and digest.
    *)
+  let split_further lst =
+    split_newline ~do_trim:false (String.concat "\n" lst)
+  in
+  let oasis_section =
+    match t.body with
+      | NoBody ->
+        []
+
+      | BodyWithDigest (d, lst) ->
+          (t.comment.of_string
+             (Printf.sprintf
+                "DO NOT EDIT (digest: %s)"
+                (Digest.to_hex d)))
+          :: (split_further lst)
+      | Body lst ->
+        split_further lst
+  in
+    List.flatten
+      [
+        t.header;
+        if t.disable_oasis_section then
+          oasis_section
+        else
+          t.comment.start :: oasis_section @ [t.comment.stop];
+        t.footer;
+      ]
+
+
+let to_file t =
   let chn_out =
     open_out_gen
       [Open_wronly; Open_creat; Open_trunc; Open_binary]
       t.perm
       t.fn
   in
-  let output_line str =
-    output_string chn_out str;
-    output_char   chn_out '\n'
-  in
-  let output_lst =
-    List.iter output_line
-  in
-
-    output_lst t.header;
-    begin
-      match t.body with
-        | NoBody ->
-            ()
-
-        | BodyWithDigest (d, lst) ->
-            output_line t.comment.start;
-            output_line
-              (t.comment.of_string
-                 (Printf.sprintf
-                    "DO NOT EDIT (digest: %s)"
-                    (Digest.to_hex d)));
-            output_lst   lst;
-            output_line  t.comment.stop
-
-        | Body lst ->
-            output_line  t.comment.start;
-            output_lst   lst;
-            output_line  t.comment.stop
-    end;
-    output_lst t.footer;
+    List.iter
+      (fun str ->
+         output_string chn_out str;
+         output_char chn_out '\n')
+      (to_string_list t);
     close_out chn_out;
     Unix.chmod t.fn t.perm
 
+
 type file_generate_change =
   | Create of host_filename
   | Change of host_filename * host_filename option
   | NoChange
 
+
 let file_rollback ~ctxt =
   function
     | Create fn ->
@@ -475,20 +492,15 @@ let file_rollback ~ctxt =
         (try Sys.remove fn with _ -> ())
 
     | Change (fn, Some bak) ->
-        begin
-          if Sys.file_exists bak then
-            begin
-              info ~ctxt (f_ "Restore file '%s' with backup file '%s'.")
-                fn bak;
-              Sys.rename bak fn
-            end
-          else
-            begin
-              warning ~ctxt
-                (f_ "Backup file '%s' disappear, cannot restore \
-                     file '%s'.")
-                bak fn
-            end
+        if Sys.file_exists bak then begin
+          info ~ctxt (f_ "Restore file '%s' with backup file '%s'.")
+            fn bak;
+          Sys.rename bak fn
+        end else begin
+          warning ~ctxt
+            (f_ "Backup file '%s' disappear, cannot restore \
+                 file '%s'.")
+            bak fn
         end
 
     | Change (fn, None) ->
@@ -499,7 +511,8 @@ let file_rollback ~ctxt =
     | NoChange ->
         ()
 
-let file_generate ~ctxt ~backup t =
+
+let file_generate ~ctxt ?(remove=false) ~backup t =
 
   (* Check that the files differ
    *)
@@ -530,20 +543,14 @@ let file_generate ~ctxt ~backup t =
     let rec backup_aux =
       function
         | ext :: tl ->
-            begin
-              let fn_backup =
-                fn ^ "." ^ ext
-              in
-                if not (Sys.file_exists fn_backup) then
-                  begin
-                    Sys.rename fn fn_backup;
-                    OASISFileUtil.cp ~ctxt fn_backup fn;
-                    fn_backup
-                  end
-                else
-                  begin
-                    backup_aux tl
-                  end
+            let fn_backup = fn ^ "." ^ ext in
+            if not (Sys.file_exists fn_backup) then begin
+              info ~ctxt (f_ "Rename '%s' to '%s'.") fn fn_backup;
+              Sys.rename fn fn_backup;
+              OASISFileUtil.cp ~ctxt fn_backup fn;
+              fn_backup
+            end else begin
+              backup_aux tl
             end
 
         | [] ->
@@ -562,96 +569,121 @@ let file_generate ~ctxt ~backup t =
     if Sys.file_exists t.fn then
       begin
         let t_org =
-          template_of_file ~ctxt ~template:false t.fn t.comment
+          template_of_file ~ctxt ~template:false t.disable_oasis_section t.fn t.comment
         in
+          (* If remove = true then backup is ignored. *)
+          if remove && t_org.header = t.header && t_org.footer = t.footer &&
+             t.body = Body [] then
+            begin
+              info ~ctxt (f_ "%s is empty - removing") t.fn;
+              if not (digest_check t_org) then
+                 warning ~ctxt (f_ "File %s has changed, doing a backup in %s")
+                   t.fn (do_backup t.fn);
+              Sys.remove t.fn;
+              NoChange
+            end
+          else
+            match t_org.body, body_has_changed t_org t with
+              | NoBody, _ -> (* No body, nothing to do *)
+                  begin
+                    NoChange
+                  end
 
-          match t_org.body, body_has_changed t_org t with
-            | NoBody, _ -> (* No body, nothing to do *)
-                begin
-                  NoChange
-                end
+              | _, true      (* Body has changed -> regenerate *)
+              | Body _, _ -> (* Missing digest -> regenerate *)
+                  begin
+                    (* Regenerate *)
+                    let () =
+                      info ~ctxt (f_ "Regenerating file %s") t.fn
+                    in
 
-            | _, true      (* Body has changed -> regenerate *)
-            | Body _, _ -> (* Missing digest -> regenerate *)
-                begin
-                  (* Regenerate *)
-                  let () =
-                    info ~ctxt (f_ "Regenerating file %s") t.fn
-                  in
-
-                  let fn_backup =
-                    (* Create a backup if required *)
-                    if not (digest_check t_org) then
-                      begin
-                        let fn_bak =
-                          do_backup t.fn
-                        in
+                    let fn_backup =
+                      (* Create a backup if required *)
+                      if not (digest_check t_org) then begin
+                        let fn_bak = do_backup t.fn in
                           warning ~ctxt
                             (f_ "File %s has changed, doing a backup in %s")
                             t.fn fn_bak;
                           Some fn_bak
-                      end
-                    else if backup then
-                      begin
+                      end else if backup then begin
                         Some (do_backup t.fn)
-                      end
-                    else
-                      None
-                  in
-                    to_file (merge t_org t);
-                    Change (t.fn, fn_backup)
+                      end else
+                        None
+                    in
+                      to_file (merge t_org t);
+                      Change (t.fn, fn_backup)
+                  end
 
+              | _, false -> (* No change *)
+                begin
+                  info ~ctxt (f_ "File %s has not changed, skipping") t.fn;
+                  NoChange
                 end
-
-            | _, false -> (* No change *)
-              begin
-                info ~ctxt (f_ "File %s has not changed, skipping") t.fn;
-                NoChange
-              end
       end
     else
       begin
-        info ~ctxt (f_ "File %s doesn't exist, creating it.") t.fn;
-        to_file t;
-        Create t.fn
-
+        if remove then begin
+          info ~ctxt (f_ "File %s doesn't exist, deletion unnecessary.") t.fn;
+          NoChange
+        end else begin
+          info ~ctxt (f_ "File %s doesn't exist, creating it.") t.fn;
+          to_file t;
+          Create t.fn
+        end
       end
 
 
-module S =
-  Map.Make (
-struct
-  type t = host_filename
-
-  let compare =
-    OASISHostPath.compare
-end)
 
 exception AlreadyExists of host_filename
 
-type templates = template S.t
 
-let empty =
-  S.empty
+module M = OASISHostPath.Map
 
-let find =
-  S.find
+type templates =
+    {
+      files: template M.t;
+      disable_oasis_section_files: OASISUnixPath.Set.t
+    }
 
-let add e t =
-  if S.mem e.fn t then
-    raise (AlreadyExists e.fn)
-  else
-    S.add e.fn e t
 
-let remove fn t =
-  S.remove fn t
+let create ~disable_oasis_section () =
+  {
+    files = M.empty;
+    disable_oasis_section_files =
+      OASISUnixPath.Set.of_list disable_oasis_section;
+  }
+
+
+let find e t =
+  M.find e t.files
+
 
 let replace e t =
-  S.add e.fn e t
+  let e =
+    if OASISUnixPath.Set.mem
+         (OASISHostPath.to_unix e.fn)
+         t.disable_oasis_section_files then
+      {e with disable_oasis_section = true}
+    else
+      e
+  in
+    {t with files = M.add e.fn e t.files}
+
+
+let add e t =
+  if M.mem e.fn t.files then
+    raise (AlreadyExists e.fn)
+  else
+    replace e t
+
+
+let remove fn t =
+  {t with files = M.remove fn t.files}
+
 
 let fold f t acc =
-  S.fold
+  M.fold
     (fun k e acc ->
        f e acc)
-    t
+    t.files
     acc
