@@ -1,7 +1,8 @@
 (******************************************************************************)
 (* OASIS: architecture for building OCaml libraries and applications          *)
 (*                                                                            *)
-(* Copyright (C) 2008-2010, OCamlCore SARL                                    *)
+(* Copyright (C) 2011-2013, Sylvain Le Gall                                   *)
+(* Copyright (C) 2008-2011, OCamlCore SARL                                    *)
 (*                                                                            *)
 (* This library is free software; you can redistribute it and/or modify it    *)
 (* under the terms of the GNU Lesser General Public License as published by   *)
@@ -19,38 +20,52 @@
 (* Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA              *)
 (******************************************************************************)
 
+
 (** Install using internal scheme
     @author Sylvain Le Gall
   *)
 
+
 open BaseEnv
 open BaseStandardVar
 open BaseMessage
 open OASISTypes
-open OASISLibrary
+open OASISFindlib
 open OASISGettext
 open OASISUtils
 
+
 let exec_hook =
   ref (fun (cs, bs, exec) -> cs, bs, exec)
 
+
 let lib_hook =
   ref (fun (cs, bs, lib) -> cs, bs, lib, [])
 
+
+let obj_hook =
+  ref (fun (cs, bs, obj) -> cs, bs, obj, [])
+
+
 let doc_hook =
   ref (fun (cs, doc) -> cs, doc)
 
+
 let install_file_ev =
   "install-file"
 
+
 let install_dir_ev =
   "install-dir"
 
+
 let install_findlib_ev =
   "install-findlib"
 
+
 let win32_max_command_line_length = 8000
 
+
 let split_install_command ocamlfind findlib_name meta files =
   if Sys.os_type = "Win32" then
     (* Arguments for the first command: *)
@@ -90,20 +105,21 @@ let split_install_command ocamlfind findlib_name meta files =
               | (firsts, others) ->
                   let cmd = args @ firsts in
                   (* Use -add for remaining commands: *)
-                  let () = 
+                  let () =
                     let findlib_ge_132 =
                       OASISVersion.comparator_apply
-                        (OASISVersion.version_of_string 
+                        (OASISVersion.version_of_string
                            (BaseStandardVar.findlib_version ()))
-                        (OASISVersion.VGreaterEqual 
+                        (OASISVersion.VGreaterEqual
                            (OASISVersion.version_of_string "1.3.2"))
                     in
                       if not findlib_ge_132 then
                         failwithf
-                          (f_ "Installing the library %s require to use the flag \
-                               '-add' of ocamlfind because the command line is too \
-                                long. This flag is only available for findlib 1.3.2. \
-                                Please upgrade findlib from %s to 1.3.2")
+                          (f_ "Installing the library %s require to use the \
+                               flag '-add' of ocamlfind because the command \
+                               line is too long. This flag is only available \
+                               for findlib 1.3.2. Please upgrade findlib from \
+                               %s to 1.3.2")
                           findlib_name (BaseStandardVar.findlib_version ())
                   in
                   let cmds = split other_args others in
@@ -114,6 +130,7 @@ let split_install_command ocamlfind findlib_name meta files =
   else
     ["install" :: findlib_name :: meta :: files]
 
+
 let install pkg argv =
 
   let in_destdir =
@@ -257,6 +274,75 @@ let install pkg argv =
           begin
             (f_data, acc)
           end
+    and files_of_object (f_data, acc) data_obj =
+      let cs, bs, obj, obj_extra =
+        !obj_hook data_obj
+      in
+        if var_choose bs.bs_install &&
+           BaseBuilt.is_built BaseBuilt.BObj cs.cs_name then
+          begin
+            let acc =
+              (* Start with acc + obj_extra *)
+              List.rev_append obj_extra acc
+            in
+            let acc =
+              (* Add uncompiled header from the source tree *)
+              let path =
+                OASISHostPath.of_unix bs.bs_path
+              in
+                List.fold_left
+                  (fun acc modul ->
+                     try
+                       List.find
+                         OASISFileUtil.file_exists_case
+                         (List.map
+                            (Filename.concat path)
+                            [modul^".mli";
+                             modul^".ml";
+                             String.uncapitalize modul^".mli";
+                             String.capitalize   modul^".mli";
+                             String.uncapitalize modul^".ml";
+                             String.capitalize   modul^".ml"])
+                       :: acc
+                     with Not_found ->
+                       begin
+                         warning
+                           (f_ "Cannot find source header for module %s \
+                                in object %s")
+                           modul cs.cs_name;
+                         acc
+                       end)
+                  acc
+                  obj.obj_modules
+            in
+
+            let acc =
+             (* Get generated files *)
+             BaseBuilt.fold
+               BaseBuilt.BObj
+               cs.cs_name
+               (fun acc fn -> fn :: acc)
+               acc
+            in
+
+            let f_data () =
+              (* Install data associated with the object *)
+              install_data
+                bs.bs_path
+                bs.bs_data_files
+                (Filename.concat
+                   (datarootdir ())
+                   pkg.name);
+              f_data ()
+            in
+
+              (f_data, acc)
+          end
+         else
+          begin
+            (f_data, acc)
+          end
+
     in
 
     (* Install one group of library *)
@@ -267,8 +353,10 @@ let install pkg argv =
           match grp with
             | Container (_, children) ->
                 data_and_files, children
-            | Package (_, cs, bs, lib, children) ->
+            | Package (_, cs, bs, `Library lib, children) ->
                 files_of_library data_and_files (cs, bs, lib), children
+            | Package (_, cs, bs, `Object obj, children) ->
+                files_of_object data_and_files (cs, bs, obj), children
         in
           List.fold_left
             install_group_lib_aux
@@ -302,7 +390,7 @@ let install pkg argv =
           begin
             let meta =
               (* Search META file *)
-              let (_, bs, _) =
+              let _, bs, _ =
                 root_lib
               in
               let res =
@@ -315,7 +403,7 @@ let install pkg argv =
                     findlib_name;
                 res
             in
-            let files = 
+            let files =
               (* Make filename shorter to avoid hitting command max line length
                * too early, esp. on Windows.
                *)
@@ -324,24 +412,24 @@ let install pkg argv =
                 let nlen = String.length n in
                   if plen <= nlen && String.sub n 0 plen = p then
                     begin
-                      let fn_sep = 
+                      let fn_sep =
                         if Sys.os_type = "Win32" then
                           '\\'
                         else
                           '/'
                       in
                       let cutpoint = plen +
-                        (if plen < nlen && n.[plen] = fn_sep then 
+                        (if plen < nlen && n.[plen] = fn_sep then
                            1
-                         else 
+                         else
                            0)
                       in
                         String.sub n cutpoint (nlen - cutpoint)
                     end
-                  else 
+                  else
                     n
               in
-                List.map (remove_prefix (Sys.getcwd ())) files 
+                List.map (remove_prefix (Sys.getcwd ())) files
             in
               info
                 (f_ "Installing findlib library '%s'")
@@ -375,7 +463,7 @@ let install pkg argv =
 
   let install_execs pkg =
     let install_exec data_exec =
-      let (cs, bs, exec) =
+      let cs, bs, exec =
         !exec_hook data_exec
       in
         if var_choose bs.bs_install &&
@@ -422,7 +510,7 @@ let install pkg argv =
 
   let install_docs pkg =
     let install_doc data =
-      let (cs, doc) =
+      let cs, doc =
         !doc_hook data
       in
         if var_choose doc.doc_install &&
@@ -458,6 +546,7 @@ let install pkg argv =
     install_execs pkg;
     install_docs  pkg
 
+
 (* Uninstall already installed data *)
 let uninstall _ argv =
   List.iter
@@ -521,16 +610,20 @@ let uninstall _ argv =
        (BaseLog.filter
           [install_file_ev;
            install_dir_ev;
-           install_findlib_ev;]))
+           install_findlib_ev]))
+
 
 (* END EXPORT *)
 
+
 open OASISPlugin
 open InternalId
 
+
 let plugin =
   `Install, name, Some version
 
+
 let init () =
   let self_id, _ =
     Install.create plugin
