@@ -1,7 +1,8 @@
 (******************************************************************************)
 (* OASIS: architecture for building OCaml libraries and applications          *)
 (*                                                                            *)
-(* Copyright (C) 2008-2010, OCamlCore SARL                                    *)
+(* Copyright (C) 2011-2013, Sylvain Le Gall                                   *)
+(* Copyright (C) 2008-2011, OCamlCore SARL                                    *)
 (*                                                                            *)
 (* This library is free software; you can redistribute it and/or modify it    *)
 (* under the terms of the GNU Lesser General Public License as published by   *)
@@ -19,15 +20,24 @@
 (* Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA              *)
 (******************************************************************************)
 
+
 (** Functions common to OCamlbuild build and doc plugin
   *)
 
+
 open OASISGettext
 open BaseEnv
 open BaseStandardVar
+open OASISTypes
+
+
+TYPE_CONV_PATH "OCamlbuildCommon"
+
+type extra_args = string list with odn
+
+
+let ocamlbuild_clean_ev = "ocamlbuild-clean"
 
-let ocamlbuild_clean_ev =
-  "ocamlbuild-clean"
 
 let ocamlbuildflags =
   var_define
@@ -35,6 +45,7 @@ let ocamlbuildflags =
     "ocamlbuildflags"
     (fun () -> "")
 
+
 (** Fix special arguments depending on environment *)
 let fix_args args extra_argv =
   List.flatten
@@ -73,6 +84,7 @@ let fix_args args extra_argv =
       Array.to_list extra_argv;
     ]
 
+
 (** Run 'ocamlbuild -clean' if not already done *)
 let run_clean extra_argv =
   let extra_cli =
@@ -92,6 +104,7 @@ let run_clean extra_argv =
                ())
       end
 
+
 (** Run ocamlbuild, unregister all clean events *)
 let run_ocamlbuild args extra_argv =
   (* TODO: enforce that target in args must be UNIX encoded i.e. toto/index.html
@@ -103,6 +116,7 @@ let run_ocamlbuild args extra_argv =
     (fun (e, d) -> BaseLog.unregister e d)
     (BaseLog.filter [ocamlbuild_clean_ev])
 
+
 (** Determine real build directory *)
 let build_dir extra_argv =
   let rec search_args dir =
@@ -116,9 +130,10 @@ let build_dir extra_argv =
   in
     search_args "_build" (fix_args [] extra_argv)
 
+
 (* END EXPORT *)
 
-open OASISTypes
+open OASISValues
 
 let fix_build_tools tool pkg =
   let fix_build_tools' sct bs =
@@ -141,6 +156,10 @@ let fix_build_tools tool pkg =
                  let bs = fix_build_tools' sct bs in
                    Library (cs, bs, lib)
 
+             | Object (cs, bs, obj) ->
+                 let bs = fix_build_tools' sct bs in
+                   Object (cs, bs, obj)
+
              | Flag _ | SrcRepo _ | Test _ | Doc _ as sct ->
                  sct
          in
@@ -158,5 +177,83 @@ struct
     *)
   let filename_concat fn1 fn2 =
     OASISUnixPath.concat (OASISUnixPath.reduce fn1) fn2
-
 end
+
+(** Check OCaml version constraint defined in _oasis. *)
+let check_ocaml_version version pkg =
+  match pkg.ocaml_version with
+    | Some ocaml_version ->
+        let min_ocaml_version = OASISVersion.version_of_string version in
+        OASISVersion.comparator_ge min_ocaml_version ocaml_version
+    | None ->
+        false
+
+
+let ocamlbuild_more_args =
+  OASISFeatures.create "ocamlbuild_more_args"
+    OASISFeatures.alpha
+    (fun () ->
+       s_ "Allow to pass arguments to ocamlbuild.")
+
+
+let ocamlbuild_supports_ocamlfind = check_ocaml_version "3.12.1"
+let ocamlbuild_supports_plugin_tags = check_ocaml_version "4.01"
+
+
+type ocamlbuild_common =
+  {
+    plugin_tags: string option;
+    extra_args: string list;
+  } with odn
+
+
+let ocamlbuild_common_generator pivot_data schm id =
+  let new_field nm = OASISSchema.new_field schm id nm in
+  let plugin_tags =
+    new_field
+      "PluginTags"
+      ~default:None
+      ~feature:ocamlbuild_more_args
+      (opt string_not_empty)
+      (fun () -> s_ "Gives the plugin tags to ocambuild through \
+                     '-plugin-tags' (OCaml >= 4.01 only)")
+      pivot_data (fun _ t -> t.plugin_tags)
+  in
+  let extra_args =
+    new_field
+      "ExtraArgs"
+      ~default:[]
+      ~feature:ocamlbuild_more_args
+      command_line_options
+      (fun () -> s_ "Gives extra arguments to ocamlbuild")
+      pivot_data (fun _ t -> t.extra_args)
+  in
+    fun data ->
+      {
+        extra_args = extra_args data;
+        plugin_tags = plugin_tags data;
+      }
+
+
+let extra_args_ocamlbuild_common ~ctxt pkg t =
+  let extra_args =
+    if t.plugin_tags <> None &&
+       not (ocamlbuild_supports_plugin_tags pkg) then begin
+      OASISMessage.error
+        ~ctxt:ctxt
+        (f_ "'XOCamlbuildPluginTags' in only available for OCaml >= 4.01. \
+             Please restrict your requirements with 'OCamlVersion: >= 4.01'");
+      t.extra_args
+    end else begin
+      match t.plugin_tags with
+        | Some tags -> "-plugin-tags" :: ("'" ^ tags ^ "'") :: t.extra_args
+        | None -> t.extra_args
+    end
+  in
+  let extra_args =
+    if ocamlbuild_supports_ocamlfind pkg then
+      "-use-ocamlfind" :: extra_args
+    else
+      extra_args
+  in
+  extra_args
