@@ -1,7 +1,8 @@
 (******************************************************************************)
 (* OASIS: architecture for building OCaml libraries and applications          *)
 (*                                                                            *)
-(* Copyright (C) 2008-2010, OCamlCore SARL                                    *)
+(* Copyright (C) 2011-2013, Sylvain Le Gall                                   *)
+(* Copyright (C) 2008-2011, OCamlCore SARL                                    *)
 (*                                                                            *)
 (* This library is free software; you can redistribute it and/or modify it    *)
 (* under the terms of the GNU Lesser General Public License as published by   *)
@@ -19,23 +20,30 @@
 (* Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA              *)
 (******************************************************************************)
 
+
 (** Build using ocamlbuild
     @author Sylvain Le Gall
   *)
 
+
 open OASISTypes
 open OASISGettext
 open OASISUtils
+open OASISString
 open BaseEnv
 open OCamlbuildCommon
 open BaseStandardVar
 open BaseMessage
 
+
+TYPE_CONV_PATH "OCamlbuildPlugin"
+
+
 let cond_targets_hook =
   ref (fun lst -> lst)
 
-let build pkg argv =
 
+let build extra_args pkg argv =
   (* Return the filename in build directory *)
   let in_build_dir fn =
     Filename.concat
@@ -60,18 +68,6 @@ let build pkg argv =
                      (cs, bs, lib)
                  in
 
-                 let ends_with nd fn =
-                   let nd_len =
-                     String.length nd
-                   in
-                     (String.length fn >= nd_len)
-                     &&
-                     (String.sub
-                        fn
-                        (String.length fn - nd_len)
-                        nd_len) = nd
-                 in
-
                  let tgts =
                    List.flatten
                      (List.filter
@@ -79,11 +75,11 @@ let build pkg argv =
                         (List.map
                            (List.filter
                               (fun fn ->
-                               ends_with ".cma" fn
-                               || ends_with ".cmxs" fn
-                               || ends_with ".cmxa" fn
-                               || ends_with (ext_lib ()) fn
-                               || ends_with (ext_dll ()) fn))
+                               ends_with ~what:".cma" fn
+                               || ends_with ~what:".cmxs" fn
+                               || ends_with ~what:".cmxa" fn
+                               || ends_with ~what:(ext_lib ()) fn
+                               || ends_with ~what:(ext_dll ()) fn))
                            unix_files))
                  in
 
@@ -96,6 +92,35 @@ let build pkg argv =
                            cs.cs_name
                end
 
+           | Object (cs, bs, obj) when var_choose bs.bs_build ->
+               begin
+                 let evs, unix_files =
+                   BaseBuilt.of_object
+                     in_build_dir_of_unix
+                     (cs, bs, obj)
+                 in
+
+                 let tgts =
+                   List.flatten
+                     (List.filter
+                        (fun l -> l <> [])
+                        (List.map
+                           (List.filter
+                              (fun fn ->
+                               ends_with ".cmo" fn
+                               || ends_with ".cmx" fn))
+                           unix_files))
+                 in
+
+                   match tgts with
+                     | _ :: _ ->
+                         (evs, tgts) :: acc
+                     | [] ->
+                         failwithf
+                           (f_ "No possible ocamlbuild targets for object %s")
+                           cs.cs_name
+               end
+
            | Executable (cs, bs, exec) when var_choose bs.bs_build ->
                begin
                  let evs, unix_exec_is, unix_dll_opt =
@@ -111,12 +136,13 @@ let build pkg argv =
                         (OASISUnixPath.chop_extension
                            exec.exec_main_is))^ext
                    in
-                   let evs = 
+                   let evs =
                      (* Fix evs, we want to use the unix_tgt, without copying *)
                      List.map
                        (function
                           | BaseBuilt.BExec, nm, lst when nm = cs.cs_name ->
-                              BaseBuilt.BExec, nm, [[in_build_dir_of_unix unix_tgt]]
+                              BaseBuilt.BExec, nm,
+                              [[in_build_dir_of_unix unix_tgt]]
                           | ev ->
                               ev)
                        evs
@@ -138,7 +164,7 @@ let build pkg argv =
                    acc
                end
 
-           | Library _ | Executable _ | Test _
+           | Library _ | Object _ | Executable _ | Test _
            | SrcRepo _ | Flag _ | Doc _ ->
                acc)
       []
@@ -152,26 +178,22 @@ let build pkg argv =
       (fun fns ->
          if not (List.exists OASISFileUtil.file_exists_case fns) then
            failwithf
-             (f_ "No one of expected built files %s exists")
-             (String.concat (s_ ", ") (List.map (Printf.sprintf "'%s'") fns)))
+             (fn_
+                "Expected built file %s doesn't exist."
+                "None of expected built files %s exists."
+                (List.length fns))
+             (String.concat (s_ " or ") (List.map (Printf.sprintf "'%s'") fns)))
       lst;
       (BaseBuilt.register bt bnm lst)
   in
 
-  let cond_targets =
-    (* Run the hook *)
-    !cond_targets_hook cond_targets
-  in
+  (* Run the hook *)
+  let cond_targets = !cond_targets_hook cond_targets in
 
-    (* Run a list of target... *)
-    run_ocamlbuild 
-      (List.flatten 
-         (List.map snd cond_targets))
-      argv;
-    (* ... and register events *)
-    List.iter
-      check_and_register
-      (List.flatten (List.map fst cond_targets))
+  (* Run a list of target... *)
+  run_ocamlbuild (List.flatten (List.map snd cond_targets) @ extra_args) argv;
+  (* ... and register events *)
+  List.iter check_and_register (List.flatten (List.map fst cond_targets))
 
 
 let clean pkg extra_args  =
@@ -187,8 +209,10 @@ let clean pkg extra_args  =
            ())
     pkg.sections
 
+
 (* END EXPORT *)
 
+
 open OASISFileTemplate
 open OASISUtils
 open OASISMessage
@@ -196,27 +220,44 @@ open OASISGettext
 open ODN
 open OASISPlugin
 open OASISTypes
-open OASISValues
+open OASISSchema
 open MyOCamlbuildBase
 open Ocamlbuild_plugin
 open OCamlbuildId
 
+
 let plugin =
   `Build, name, Some version
 
+
 let self_id, all_id =
   Build.create plugin
 
+
+let pure_interface_test =
+  OASISFeatures.package_test
+    (OASISFeatures.create "pure_interface" ~plugin
+       OASISFeatures.alpha
+       (fun () ->
+          s_ "Allow to have module with only .mli file."))
+
+
+let pivot_data =
+  data_new_property plugin
+
+
 let only_h_files lst =
   List.filter
     (fun fn -> OASISUnixPath.check_extension fn "h")
     lst
 
+
 let only_c_files lst =
   List.filter
     (fun fn -> OASISUnixPath.check_extension fn "c")
     lst
 
+
 let add_tags tag_t tgts tags =
   let quote_target fn =
     let test_char =
@@ -245,6 +286,7 @@ let add_tags tag_t tgts tags =
 let prepend_bs_path bs fn =
   Tag.filename_concat bs.bs_path fn
 
+
 let bs_paths bs files =
   let subdirs =
     List.rev_map OASISUnixPath.dirname
@@ -252,16 +294,18 @@ let bs_paths bs files =
   in
     (* Unique elements *)
     SetString.elements
-      (set_string_of_list
+      (SetString.of_list
          (List.rev_map OASISUnixPath.reduce (bs.bs_path :: subdirs)))
 
-let bs_tags pkg sct cs bs src_dirs src_internal_dirs link_tgt ctxt tag_t myocamlbuild_t =
+
+let bs_tags pkg sct cs bs src_dirs src_internal_dirs link_tgt ctxt tag_t
+      myocamlbuild_t =
 
   let link_pkg =
     (* Only link findlib package with executable *)
     match sct with
       | Executable _ -> true
-      | Library _ | Flag _ | Test _ | SrcRepo _ | Doc _ -> false
+      | Library _ | Object _ | Flag _ | Test _ | SrcRepo _ | Doc _ -> false
   in
 
   let src_tgts =
@@ -459,21 +503,28 @@ let bs_tags pkg sct cs bs src_dirs src_internal_dirs link_tgt ctxt tag_t myocaml
     let mp =
       OASISBuildSection.transitive_build_depends pkg
     in
+    let supports_ocamlfind = ocamlbuild_supports_ocamlfind pkg in
       add_tags
         tag_t
         (if link_pkg then
            link_tgt :: src_tgts
          else
            src_tgts)
-        (List.fold_left
-           (fun acc ->
-              function
-                | FindlibPackage (findlib_pkg, _) ->
-                    ("pkg_"^findlib_pkg) :: acc
-                | InternalLibrary nm ->
-                    ("use_"^nm) :: acc)
-           []
-           (OASISSection.MapSection.find sct mp))
+        (List.sort String.compare
+          (List.fold_left
+             (fun acc ->
+                function
+                  | FindlibPackage (findlib_pkg, _) ->
+                      (if supports_ocamlfind then
+                         ("package("^findlib_pkg^")")
+                       else
+                         ("pkg_"^findlib_pkg)
+                      )
+                      :: acc
+                  | InternalLibrary nm ->
+                      ("use_"^nm) :: acc)
+             []
+             (OASISSection.MapSection.find sct mp)))
   in
 
   let ctxt =
@@ -488,36 +539,43 @@ let bs_tags pkg sct cs bs src_dirs src_internal_dirs link_tgt ctxt tag_t myocaml
 
     ctxt, tag_t, myocamlbuild_t
 
-module MapDirs = 
-  Map.Make 
-    (struct 
-       type t = [`Library of string | `Executable of string]
-       let compare t1 t2 = 
-         match t1, t2 with 
-           | `Library _, `Executable _ ->
+
+module MapDirs =
+  Map.Make
+    (struct
+       type t = [`Library of string | `Object of string | `Executable of string]
+       let compare t1 t2 =
+         match t1, t2 with
+           | `Library _, `Executable _
+           | `Library _, `Object _
+           | `Object _, `Executable _ ->
                -1
-           | `Executable _, `Library _ ->
+           | `Object _, `Library _
+           | `Executable _, `Library _
+           | `Executable _, `Object _ ->
                1
-           | `Library s1, `Library s2 
+           | `Library s1, `Library s2
+           | `Object s1, `Object s2
            | `Executable s1, `Executable s2 ->
                String.compare s1 s2
      end)
 
-let compute_map_dirs pkg = 
+
+let compute_map_dirs pkg =
   let add k dirs internal_dirs mp =
-    let src_dirs, src_internal_dirs = 
-      try 
+    let src_dirs, src_internal_dirs =
+      try
         MapDirs.find k mp
       with Not_found ->
         SetString.empty, SetString.empty
     in
-    let add_dirs st dirs = SetString.union st (set_string_of_list dirs) in
-      MapDirs.add k 
-        (add_dirs src_dirs dirs, 
-         add_dirs src_internal_dirs internal_dirs) 
+    let add_dirs st dirs = SetString.add_list st dirs in
+      MapDirs.add k
+        (add_dirs src_dirs dirs,
+         add_dirs src_internal_dirs internal_dirs)
         mp
   in
-  let map_dirs = 
+  let map_dirs =
     List.fold_left
       (fun mp ->
          function
@@ -529,6 +587,14 @@ let compute_map_dirs pkg =
                  (bs_paths bs lib.lib_internal_modules)
                  mp
 
+           | Object (cs, bs, obj) ->
+               add (`Object cs.cs_name)
+                 (* All paths accessed from within the (potentially packed)
+                    object *)
+                 (bs_paths bs obj.obj_modules)
+                 [] (* No internal paths *)
+                 mp
+
            | Executable (cs, bs, exec) ->
                add (`Executable cs.cs_name)
                  (bs_paths bs [exec.exec_main_is])
@@ -546,30 +612,32 @@ let compute_map_dirs pkg =
          SetString.elements (SetString.diff src_internal_dirs src_dirs))
       map_dirs
 
+
 let compute_includes map_dirs pkg =
-  let add_includes dir set_dirs includes = 
+  let add_includes dir set_dirs includes =
     (* Not self-dependent *)
-    let set_dirs = SetString.diff set_dirs (SetString.singleton dir) in 
-    let pre_dirs = 
-      try 
-        MapString.find dir includes 
+    let set_dirs = SetString.diff set_dirs (SetString.singleton dir) in
+    let pre_dirs =
+      try
+        MapString.find dir includes
       with Not_found ->
         SetString.empty
     in
       MapString.add dir (SetString.union set_dirs pre_dirs) includes
   in
 
-  let add_map_dirs k bs includes = 
-    let dep_dirs = 
+  let add_map_dirs k bs includes =
+    let dep_dirs =
       (* Source dirs of dependent libraries *)
-      List.fold_left 
+      List.fold_left
         (fun set ->
            function
              | InternalLibrary nm ->
-                let src_dirs, _ = 
-                   MapDirs.find (`Library nm) map_dirs 
+                let src_dirs, _ =
+                   try MapDirs.find (`Library nm) map_dirs
+                   with Not_found -> MapDirs.find (`Object nm) map_dirs
                  in
-                   SetString.union set (set_string_of_list src_dirs)
+                   SetString.add_list set src_dirs
 
              | FindlibPackage _ ->
                  set)
@@ -579,18 +647,16 @@ let compute_includes map_dirs pkg =
     let self_dirs =
       (* Source dirs *)
       let src_dirs, src_internal_dirs = MapDirs.find k map_dirs in
-        SetString.union 
-          (set_string_of_list src_dirs)
-          (set_string_of_list src_internal_dirs)
+        SetString.add_list (SetString.of_list src_dirs) src_internal_dirs
     in
     let all_dirs = SetString.union dep_dirs self_dirs in
-    let all_dirs = 
+    let all_dirs =
       (* No need to include the current dir. *)
       SetString.filter
         (fun dn -> not (OASISUnixPath.is_current_dir dn))
         all_dirs
     in
-        
+
       (* All self_dirs depends on all_dirs *)
       SetString.fold
         (fun dir includes ->
@@ -599,12 +665,14 @@ let compute_includes map_dirs pkg =
         includes
   in
 
-  let includes = 
+  let includes =
     List.fold_left
       (fun includes ->
          function
            | Library (cs, bs, _) ->
                add_map_dirs (`Library cs.cs_name) bs includes
+           | Object (cs, bs, _) ->
+               add_map_dirs (`Object cs.cs_name) bs includes
            | Executable (cs, bs, _) ->
                add_map_dirs (`Executable cs.cs_name) bs includes
            | Flag _ | SrcRepo _ | Test _ | Doc _ ->
@@ -621,9 +689,26 @@ let compute_includes map_dirs pkg =
       includes
       []
 
+
+(* Check if the given files list only contains .mli. *)
+let is_pure_interface (base_fn, fn_lst) =
+  let rec is_pure_interface_aux =
+    (* TODO: this needs to be refine because sometime we don't have the .ml file
+     * because it is generated (BaseData.ml) but we have the .mli.
+     *)
+    function
+      | [fn] ->
+          OASISString.ends_with ~what:".mli" fn
+      | fn :: tl ->
+          OASISString.ends_with ~what:".mli" fn && is_pure_interface_aux tl
+      | [] -> false
+  in
+    is_pure_interface_aux fn_lst
+
+
 let add_ocamlbuild_files ctxt pkg =
 
-  let map_dirs = 
+  let map_dirs =
     compute_map_dirs pkg
   in
 
@@ -636,7 +721,7 @@ let add_ocamlbuild_files ctxt pkg =
                  (* Extract content for libraries *)
 
                  (* All paths of the library *)
-                 let src_dirs, src_internal_dirs = 
+                 let src_dirs, src_internal_dirs =
                    MapDirs.find (`Library cs.cs_name) map_dirs
                  in
 
@@ -660,6 +745,32 @@ let add_ocamlbuild_files ctxt pkg =
                    (Printf.sprintf "# Library %s" cs.cs_name) :: tag_t
                  in
 
+                 (* Sources of the library. *)
+                 let sources =
+                   OASISLibrary.source_unix_files
+                     ~ctxt:ctxt.ctxt
+                     (cs, bs, lib)
+                     (fun ufn ->
+                        OASISFileUtil.file_exists_case
+                          (OASISHostPath.of_unix ufn))
+                 in
+
+                 let intf_module_list, impl_module_list =
+                   let to_module (base_fn, _) =
+                     String.capitalize (Filename.basename base_fn)
+                   in
+                   let intf_module_list =
+                     if pure_interface_test pkg then
+                       List.map to_module
+                         (List.filter is_pure_interface sources)
+                     else
+                       []
+                   in
+                     List.partition
+                       (fun modul -> List.mem modul intf_module_list)
+                       (lib.lib_modules @ lib.lib_internal_modules)
+                 in
+
                  (* Add dependency of cmxs to their own library: used
                     at link time when there is C code *)
                  let tag_t =
@@ -673,10 +784,141 @@ let add_ocamlbuild_files ctxt pkg =
 
                  let tag_t =
                    if lib.lib_pack then
+                     add_tags
+                       tag_t
+                       (List.rev_map
+                          (fun (base_fn, _) ->
+                             OASISUnixPath.add_extension base_fn "cmx")
+                          sources)
+                       ["for-pack("^String.capitalize cs.cs_name^")"]
+                   else
+                     tag_t
+                 in
+
+                 let ctxt, tag_t, myocamlbuild_t =
+                   bs_tags
+                     pkg sct cs bs
+                     src_dirs
+                     src_internal_dirs
+                     target_lib
+                     ctxt
+                     tag_t
+                     myocamlbuild_t
+                 in
+
+                 let myocamlbuild_t =
+                   {myocamlbuild_t with
+                        lib_ocaml =
+                          (cs.cs_name,
+                           List.filter
+                             (fun fn -> not (OASISUnixPath.is_current fn))
+                             src_dirs,
+                           intf_module_list)
+                          :: myocamlbuild_t.lib_ocaml}
+                 in
+
+                 let () =
+                   if lib.lib_modules = [] then
+                     warning
+                       ~ctxt:ctxt.ctxt
+                       (f_ "No exported module defined for library %s")
+                       cs.cs_name;
+                 in
+
+                 let ctxt =
+                   (* Generate .mllib *)
+                   let fn_base = prepend_bs_path bs cs.cs_name in
+                   let mllib = OASISHostPath.add_extension fn_base "mllib" in
+                   let mldylib =
+                     OASISHostPath.add_extension fn_base "mldylib"
+                   in
+                   let mlpack = OASISHostPath.add_extension fn_base "mlpack" in
+                   let mllib_template_lines =
+                     (* mllib contains either the name of the pack or the list
+                      * of modules. *)
+                     if lib.lib_pack then
+                       [ String.capitalize cs.cs_name ]
+                     else
+                       impl_module_list
+                   in
+                   let ctxt =
+                     add_file
+                       (template_make mllib
+                          comment_ocamlbuild
+                          []
+                          mllib_template_lines
+                          [])
+                       ctxt
+                   in
+                   let ctxt =
+                     add_file
+                       (template_make mldylib
+                          comment_ocamlbuild
+                          []
+                          mllib_template_lines
+                          [])
+                       ctxt
+                   in
+                   if lib.lib_pack then begin
+                     (* generate .mlpack for packed libraries *)
+                     add_file
+                       (template_make mlpack
+                          comment_ocamlbuild
+                          []
+                          impl_module_list
+                          [])
+                       ctxt
+                   end else begin
+                     (* make sure there is no conflicting mlpack file *)
+                     {ctxt with
+                          other_actions =
+                            (fun ()->
+                              if OASISFileUtil.file_exists_case mlpack then
+                                OASISMessage.error ~ctxt:ctxt.ctxt
+                                  (f_ "Conflicting file '%s' and '%s' \
+                                      exists, remove '%s'.")
+                                  mllib mlpack mlpack)
+                          :: ctxt.other_actions}
+                   end
+                 in
+                 ctxt, tag_t, myocamlbuild_t
+               end
+
+           | Object (cs, bs, obj) as sct ->
+               begin
+                 (* Extract content for objects *)
+
+                 (* All paths of the library *)
+                 let src_dirs, src_internal_dirs =
+                   MapDirs.find (`Object cs.cs_name) map_dirs
+                 in
+
+                 (* Generated library *)
+                 let target_lib =
+                   let ext =
+                     match bs.bs_compiled_object with
+                       | Best ->
+                           "{cmo,cmx}"
+                       | Byte ->
+                           "cmo"
+                       | Native ->
+                           "cmx"
+                   in
+                     prepend_bs_path bs
+                       (OASISUnixPath.add_extension cs.cs_name ext)
+                 in
+
+                 (* Start comment *)
+                 let tag_t =
+                   (Printf.sprintf "# Object %s" cs.cs_name) :: tag_t
+                 in
+
+                 let tag_t =
+                   if List.length obj.obj_modules <> 1 then
                      let base_sources =
-                       OASISLibrary.source_unix_files
+                       OASISObject.source_unix_files
                          ~ctxt:ctxt.ctxt
-                         (cs, bs, lib)
+                         (cs, bs, obj)
                          (fun ufn ->
                             OASISFileUtil.file_exists_case
                               (OASISHostPath.of_unix ufn))
@@ -709,53 +951,49 @@ let add_ocamlbuild_files ctxt pkg =
                           (cs.cs_name,
                            List.filter
                              (fun fn -> not (OASISUnixPath.is_current fn))
-                             src_dirs) :: myocamlbuild_t.lib_ocaml}
+                             src_dirs,
+                           []) :: myocamlbuild_t.lib_ocaml}
                  in
 
                  let () =
-                   if lib.lib_modules = [] then
+                   if obj.obj_modules = [] then
                      warning
                        ~ctxt:ctxt.ctxt
-                       (f_ "No exported module defined for library %s")
+                       (f_ "No exported module defined for object %s")
                        cs.cs_name;
                  in
 
                  let ctxt =
-                   (* Generate .mllib or .mlpack files *)
-                   let extension, not_extension =
-                     if lib.lib_pack then
-                       "mlpack", "mllib"
-                     else
-                       "mllib", "mlpack"
-                   in
-                   let fn_base = prepend_bs_path bs cs.cs_name in
-                   let fn =
-                     OASISHostPath.add_extension fn_base extension
-                   in
-                   let not_fn =
-                     OASISHostPath.add_extension fn_base not_extension
-                   in
-                   let ctxt =
-                     add_file
-                       (template_make
-                          fn
-                          comment_ocamlbuild
-                          []
-                          (lib.lib_modules @ lib.lib_internal_modules)
-                          [])
-                       ctxt
-                   in
-                     {ctxt with
-                          other_actions =
-                            (fun ()->
-                               if OASISFileUtil.file_exists_case not_fn then
-                                 OASISMessage.error ~ctxt:ctxt.ctxt
-                                   (f_ "Conflicting file '%s' and '%s' \
-                                      exists, remove '%s'.")
-                                   fn not_fn not_fn)
-                          :: ctxt.other_actions}
+                   match obj.obj_modules with
+                     | [ m ] -> ctxt
+                     | _ -> (* generate mlpack file *)
+                         let fn_base = prepend_bs_path bs cs.cs_name in
+                         let fn =
+                           OASISHostPath.add_extension fn_base "mlpack"
+                         and not_fn =
+                           OASISHostPath.add_extension fn_base "ml"
+                         in
+                         let ctxt =
+                           add_file
+                             (template_make
+                                fn
+                                comment_ocamlbuild
+                                []
+                                obj.obj_modules
+                                [])
+                             ctxt
+                         in
+                           {ctxt with
+                                other_actions =
+                                  (fun () ->
+                                     if OASISFileUtil.file_exists_case not_fn
+                                     then
+                                       OASISMessage.error ~ctxt:ctxt.ctxt
+                                         (f_ "Conflicting file '%s' and '%s' \
+                                            exists, remove '%s'.")
+                                         fn not_fn not_fn)
+                                :: ctxt.other_actions}
                  in
-
                    ctxt, tag_t, myocamlbuild_t
                end
 
@@ -789,7 +1027,7 @@ let add_ocamlbuild_files ctxt pkg =
                    bs_tags
                      pkg sct cs bs
                      src_dirs
-                     src_internal_dirs 
+                     src_internal_dirs
                      target_exec
                      ctxt
                      tag_t
@@ -812,13 +1050,14 @@ let add_ocamlbuild_files ctxt pkg =
       pkg.sections
   in
 
-  (* Filter duplicate and reverse content in tag_t *)
+  (* Filter duplicate and reverse content in tag_t. *)
   let tag_t =
     snd
       (List.fold_left
          (fun (prev_tag, acc) tag ->
-            if (String.length tag > 0 && tag.[0] = '#') || (* Don't remove comment *)
-               not (SetString.mem tag prev_tag) then (* Remove already seen tag *)
+            (* Don't remove comment but remove already seen tag. *)
+            if (String.length tag > 0 && tag.[0] = '#') ||
+               not (SetString.mem tag prev_tag) then
               (
                 SetString.add tag prev_tag,
                 tag :: acc
@@ -839,10 +1078,10 @@ let add_ocamlbuild_files ctxt pkg =
     }
   in
 
-  let tag_t = 
-    "# Ignore VCS directories, you can use the same kind of rule outside "
+  let tag_t =
+    "# Ignore VCS directories, you can use the same kind of rule outside"
     ::
-    "# OASIS_START/STOP if you want to exclude directories that contains "
+    "# OASIS_START/STOP if you want to exclude directories that contains"
     ::
     "# useless stuff for the build process"
     ::
@@ -893,23 +1132,33 @@ let add_ocamlbuild_files ctxt pkg =
 
     ctxt
 
+
+let generator =
+  ocamlbuild_common_generator pivot_data OASISPackage.schema all_id
+
+let doit ctxt pkg =
+  let extra_args =
+    extra_args_ocamlbuild_common ~ctxt:ctxt.ctxt pkg
+      (generator pkg.schema_data)
+  in
+  let ctxt = add_ocamlbuild_files ctxt pkg in
+    ctxt,
+    {
+      chng_moduls       = [OCamlbuildData.ocamlbuildsys_ml];
+      chng_main         = ODNFunc.func_with_arg build
+                            "OCamlbuildPlugin.build"
+                            extra_args odn_of_extra_args;
+      chng_clean        = Some (ODNFunc.func clean "OCamlbuildPlugin.clean");
+      chng_distclean    = None;
+    }
+
+
 let qstrt_completion pkg =
   fix_build_tools (ExternalTool "ocamlbuild") pkg
 
-let init () =
-  let doit ctxt pkg =
-    let ctxt =
-      add_ocamlbuild_files ctxt pkg
-    in
 
-      ctxt,
-      {
-        chng_moduls       = [OCamlbuildData.ocamlbuildsys_ml];
-        chng_main         = ODNFunc.func build "OCamlbuildPlugin.build";
-        chng_clean        = Some (ODNFunc.func clean "OCamlbuildPlugin.clean");
-        chng_distclean    = None;
-      }
-  in
-    OCamlbuildId.init ();
-    Build.register_act self_id doit;
-    register_quickstart_completion all_id qstrt_completion
+let init () =
+  OCamlbuildId.init ();
+  Build.register_act self_id doit;
+  register_quickstart_completion all_id qstrt_completion;
+  register_generator_package all_id pivot_data generator
